################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Pipeline.py - Tools for ruffus pipelines
========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

The :mod:`Pipeline` module contains various utility functions
for interfacing CGAT ruffus pipelines with databases and the 
cluster.

API
----

'''
import os, sys, re, subprocess, optparse, stat, tempfile
import time, random, inspect, types, multiprocessing
import logging, collections, shutil, glob, gzip
import ConfigParser
import Database

# talking to a cluster
import drmaa
# talking to mercurial
import hgapi

# CGAT specific options - later to be removed
from Local import *

from ruffus import *

# use threading instead of multiprocessing
from multiprocessing.pool import ThreadPool
# note that threading can cause problems with rpy.
task.Pool = ThreadPool

import logging as L
import Experiment as E
import IOTools

# global options and arguments
GLOBAL_OPTIONS, GLOBAL_ARGS = None, None

# global drmaa session
GLOBAL_SESSION = None

class PipelineError( Exception ): pass

# Sort out the important paths
LIB_DIR=os.path.dirname( __file__ )
ROOT_DIR=os.path.dirname( LIB_DIR )
SCRIPTS_DIR=os.path.join( ROOT_DIR, "scripts" )
# if Pipeline.py is from an installed version,
# scripts are located in the "bin" directory.
if not os.path.exists( SCRIPTS_DIR):
    SCRIPTS_DIR = os.path.join(sys.exec_prefix, "bin")
    
# possible to use defaultdict, but then statements will
# fail on execution if a parameter does not exists, and not
# while building the statement. Hence, use dict.
PARAMS = { 
    'scriptsdir' : SCRIPTS_DIR,
    'toolsdir' : SCRIPTS_DIR,
    'cmd-farm' : """%s/farm.py 
                --method=drmaa 
                --cluster-priority=-10 
		--cluster-queue=all.q 
		--cluster-num-jobs=100 
                --bashrc=%s/bashrc.cgat
		--cluster-options="" """ % (SCRIPTS_DIR,SCRIPTS_DIR),
    'cmd-sql' : """sqlite3 -header -csv -separator $'\\t' """,
    'cmd-run' : """%s/run.py""" % SCRIPTS_DIR
    }

# path until parameter sharing is resolved between CGAT module
# and the pipelines module.
import Local
Local.PARAMS = PARAMS

hostname = os.uname()[0]

CONFIG = {}

# local temporary directory to use
TMPDIR = '/scratch'

def configToDictionary( config ):

    p = {}
    for section in config.sections():
        for key,value in config.items( section ):
            v = IOTools.convertValue( value )
            p["%s_%s" % (section,key)] = v
            if section in ( "general", "DEFAULT" ):
                p["%s" % (key)] = v
               
    for key, value in config.defaults().iteritems():
        p["%s" % (key)] =  IOTools.convertValue( value )
        
    return p

def getParameters( filenames = ["pipeline.ini",],
                   defaults = None ):
    '''read a config file and return as a dictionary.

    Sections and keys are combined with an underscore. If
    a key without section does not exist, it will be added 
    plain.

    For example::

       [general]
       input=input1.file

       [special]
       input=input2.file

    will be entered as { 'general_input' : "input1.file",
    'input: "input1.file", 'special_input' : "input2.file" }

    This function also updates the module-wide parameter map.
    
    The section [DEFAULT] is equivalent to [general].
    '''

    global CONFIG

    CONFIG = ConfigParser.ConfigParser()
    CONFIG.read( filenames )

    p = configToDictionary( CONFIG )
    if defaults: PARAMS.update( defaults )
    PARAMS.update( p )

    return PARAMS

def loadParameters( filenames  ):
    '''load parameters from a tuple of filenames.
    
    Parameters are processed in the same way as
    getParameters, but the global parameter dictionary
    is not updated.
    '''
    config = ConfigParser.ConfigParser()
    config.read( filenames )

    p = configToDictionary( config )
    return p

def substituteParameters( **kwargs ):
    '''return a local PARAMS dictionary.

    Options in ``**kwargs`` substitute default
    values in PARAMS.

    Finally, task specific configuration values 
    are inserted.
    '''

    # build parameter dictionary
    # note the order of addition to make sure that kwargs takes precedence
    local_params = dict(PARAMS.items() + kwargs.items())

    if "outfile" in local_params:
        # replace specific parameters with task (outfile) specific parameters
        outfile = local_params["outfile"]
        for k in local_params.keys():
            if k.startswith(outfile):
                p = k[len(outfile)+1:]
                if p not in local_params:
                    raise KeyError( "task specific parameter '%s' does not exist for '%s' " % (p,k))
                E.debug( "substituting task specific parameter for %s: %s = %s" % (outfile,p,local_params[k] ) )
                local_params[p] = local_params[k]

    return local_params

def checkFiles( filenames ):
    """check for the presence/absence of files"""
    
    missing = []
    for filename in filenames:
        if not os.path.exists( filename ):
            missing.append( filename )

    if missing:
        raise ValueError( "missing files: %s" % ",".join(missing) )

def which( filename ):

    if not os.environ.has_key('PATH') or os.environ['PATH'] == '':
        p = os.defpath
    else:
        p = os.environ['PATH']

    pathlist = p.split (os.pathsep)

    for path in pathlist:
        f = os.path.join(path, filename)
        if os.access(f, os.X_OK):
            return f
    return None

def clone( infile, outfile ):
    '''create a clone of ``infile`` named ``outfile``
    by creating a soft-link.
    '''
    # link via relative paths, otherwise it 
    # fails if infile and outfile are in different
    # directories or in a subdirectory
    if os.path.dirname(infile) != os.path.dirname(outfile):
        relpath = os.path.relpath( os.path.dirname(infile), os.path.dirname(outfile) )
    else:
        relpath = "."
    target = os.path.join( relpath, os.path.basename( infile ) )

    try:
        os.symlink( target, outfile )
    except OSError:
        pass
    
def touch( filename, times = None ):
    '''update/create a sentinel file.

    Compressed files (ending in .gz) are created
    as empty 'gzip' files, i.e., with a header.
    '''
    existed = os.path.exists(filename)
    fhandle = file(filename, 'a')

    if filename.endswith (".gz") and not existed:
        # this will automatically add a gzip header
        fhandle = gzip.GzipFile( filename, fileobj = fhandle )
        
    try:
        os.utime(filename, times)
    finally:
        fhandle.close()
    
def getTempFilename( dir = TMPDIR ):
    '''get a temporary filename.

    The caller needs to delete it.
    '''
    tmpfile = tempfile.NamedTemporaryFile(dir=dir,delete=False, prefix = "ctmp" )
    tmpfile.close()
    return tmpfile.name

def getTempFile( dir = TMPDIR ):
    '''get a temporary file.

    The caller needs to delete it.
    '''
    return tempfile.NamedTemporaryFile(dir=dir, delete=False, prefix = "ctmp" )

def getTempDir( dir = TMPDIR ):
    return tempfile.mkdtemp( dir=dir, prefix = "ctmp" )

def checkExecutables( filenames ):
    """check for the presence/absence of executables in the path"""
    
    missing = []

    for filename in filenames:
        if not which( filename ):
            missing.append( filename )

    if missing:
        raise ValueError( "missing executables: %s" % ",".join(missing) )

def checkScripts( filenames ):
    """check for the presence/absence of scripts"""
    missing = []
    for filename in filenames:
        if not os.path.exists( filename ):
            missing.append( filename )

    if missing:
        raise ValueError( "missing scripts: %s" % ",".join(missing) )

def checkParameter( key ):
    if key not in PARAMS:
        raise ValueError("need `%s` to be set" % key )

def log( loglevel, message ):
    """log message at loglevel."""
    E.log( loglevel, message )

def info( message ):
    L.info( message )

def warning( message ):
    E.warning( message )

def warn( message ):
    E.warning( message )

def debug( message ):
    L.debug( message )

def error( message ):
    E.error( message )
        
def critical( message):
    E.critical( message )

def isEmpty( filename ):
    '''return true if file *filename* is empty.
    
    If filename ends in .gz, it checks if the contents are empty
    by opening it.
    '''
    if filename.endswith(".gz"):
        n = 0
        with gzip.open(filename) as inf:
            return len(inf.read(10)) == 0
    else:
        return os.stat(filename)[6]==0

def asList( param ):
    '''return a param as a list'''
    if type(param) not in (types.ListType, types.TupleType):
        try:
            params = [x.strip() for x in param.strip().split(",")]
        except AttributeError:
            params = [param.strip()]
        return [ x for x in params if x != ""]
    else: return param

def asTuple( param ):
    '''return a param as a list'''
    return tuple(asList( param ))

def flatten(l, ltypes=(list, tuple)):
    '''flatten a nested list/tuple.'''
    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            if not l[i]:
                l.pop(i)
                i -= 1
                break
            else:
                l[i:i + 1] = l[i]
        i += 1
    return ltype(l)

def asDict( param ):
    '''return a section of configuration file as a dictionary.'''
    return dict(CONFIG.items(param))

def quote( track ):
    '''quote track such that is applicable for a table name.'''
    return re.sub( "[-(),\[\].]", "_", track)

def shellquote( statement ):
    '''shell quote a string to be used as a function argument.

    from http://stackoverflow.com/questions/967443/python-module-to-shellquote-unshellquote
    '''
    _quote_pos = re.compile('(?=[^-0-9a-zA-Z_./\n])')

    if statement:
        return _quote_pos.sub('\\\\', statement).replace('\n',"'\n'")
    else:
        return "''"

def toTable( outfile ):
    '''convert an outfile (filename) into
    a table name.

    The table name is quoted.
    '''
    assert outfile.endswith( ".load" ) 
    name = os.path.basename( outfile[:-len(".load")] )
    return quote( name )

def load( infile, 
          outfile = None, 
          options = "", 
          transpose = None,
          tablename = None):
    '''straight import from tab separated table.

    The table name is given by outfile without the
    ".load" suffix.

    If *transpose* is set, the table will be transposed before loading.
    The first column in the first row will be set to the string
    within transpose.
    '''

    if not tablename:
        tablename = toTable( outfile )
    
    statement = []
    if infile.endswith(".gz"): statement.append( "zcat %(infile)s" )
    else: statement.append( "cat %(infile)s" )

    if transpose:
        statement.append( "python %(scriptsdir)s/table2table.py --transpose --set-transpose-field=%(transpose)s" )

    statement.append('''
    python %(scriptsdir)s/csv2db.py %(csv2db_options)s 
              %(options)s 
              --table=%(tablename)s 
    > %(outfile)s
    ''')

    statement = " | ".join( statement)

    run()

def concatenateAndLoad( infiles, 
                        outfile, 
                        regex_filename = None, 
                        header = None, 
                        cat = None, 
                        titles = False, 
                        options = "" ):
    '''concatenate categorical tables and load into a database.

    Concatenation assumes that the header is the same in all files.
    The first file will be taken in completion, headers
    in other files will be removed.
    '''
    
    infiles = " ".join(infiles)

    tablename = toTable( outfile )

    passed_options = options
    load_options,options = [], []

    if regex_filename:
        options.append( '--regex-filename="%s"' % regex_filename )

    if header:
        load_options.append( "--header=%s" % header )

    if not cat:
        cat = "track"
        
    if titles == False:
        no_titles = "--no-titles"
    else: no_titles = ""

    options = " ".join(options)
    load_options = " ".join(load_options) + " " + passed_options
    statement = '''python %(scriptsdir)s/combine_tables.py
                     --cat=%(cat)s
                     %(no_titles)s
                     %(options)s
                   %(infiles)s
                   | python %(scriptsdir)s/csv2db.py %(csv2db_options)s
                      --index=track
                      --table=%(tablename)s 
                      %(load_options)s
                   > %(outfile)s'''
    run()

def mergeAndLoad( infiles, 
                  outfile, 
                  suffix = None, 
                  columns=(0,1), 
                  regex = None, 
                  row_wise = True,
                  options = "",
                  prefixes = None ):
    '''merge categorical tables and load into a database.

    The tables are merged and entered row-wise, i.e each file is 
    a row unless *row_wise* is set to False. The latter is useful if 
    histograms are being merged.

    `columns` denotes the columns to be taken. By default, the first
    two columns are taken with the first being the key. Filenames are 
    stored in a ``track`` column. Directory names are chopped off.

    If `columns` is set to None, all columns will be taken. Here, column
    names will receive a prefix. If ``prefixes`` is None, the filename
    will be added as a prefix. If ``prefixes`` is a list, the respective
    prefix will be added to each column. The length of ``prefixes`` and
    ``infiles`` need to be the same.
    '''
    if len(infiles) == 0:
        raise ValueError( "no files for merging")

    if suffix:
        header = ",".join( [ os.path.basename( snip( x, suffix) ) for x in infiles] )
    elif regex:
        header = ",".join( [ "-".join(re.search( regex, x).groups()) for x in infiles] )        
    else:
        header = ",".join( [ os.path.basename( x ) for x in infiles] )

    header_stmt = "--header=%s" % header

    if columns:
        column_filter = "| cut -f %s" % ",".join( map(str, [ x + 1 for x in columns ]))
    else:
        column_filter = ""
        if prefixes:
            assert len(prefixes) == len(infiles)
            header_stmt = "--prefixes=%s" % ",".join( prefixes)
        else:
            header_stmt = "--add-file-prefix"

    if infiles[0].endswith(".gz"):
        filenames = " ".join( [ "<( zcat %s %s )" % (x,column_filter) for x in infiles ] )
    else:
        filenames = " ".join( [ "<( cat %s %s )" % (x,column_filter) for x in infiles ] )

    tablename = toTable( outfile )

    if row_wise:
        transform = """| perl -p -e "s/bin/track/" | python %(scriptsdir)s/table2table.py --transpose""" % PARAMS
    else:
        transform = ""

    statement = """python %(scriptsdir)s/combine_tables.py
                      %(header_stmt)s
                      --skip-titles
                      --missing=0
                      --ignore-empty
                   %(filenames)s
                %(transform)s
                | python %(scriptsdir)s/csv2db.py %(csv2db_options)s
                      --index=track
                      --table=%(tablename)s 
                      %(options)s
                > %(outfile)s
            """
    run()

def createView( dbhandle, tables, tablename, outfile, view_type = "TABLE",
                ignore_duplicates = True ):
    '''create a view in database for tables.

    Tables should be a list of tuples. Each tuple
    should contain the name of a table and the field
    to join with the first table.
    
    If ignore duplicates is set to False, duplicate column
    names will be added with the tablename as prefix. The
    default is to ignore.

    For example::
       tables = ("reads_summary", "track",
                 "bam_stats", "track",
                 "context_stats", "track",
                 "picard_stats_alignment_summary_metrics", "track" )

    view_type 
       type of view. If a view is to be created across multiple database,
       use "TABLE", otherwise, use "VIEW"

    '''

    Database.executewait( dbhandle, "DROP %(view_type)s IF EXISTS %(tablename)s" % locals() )

    tracks, columns = [], []
    tablenames = [ x[0] for x in tables ]
    for table, track in tables:
        d = Database.executewait( dbhandle,"SELECT COUNT(DISTINCT %s) FROM %s" % (track,table))
        tracks.append( d.fetchone()[0] )
        columns.append( [x.lower() for x in Database.getColumnNames( dbhandle, table ) if x != track ] )

    E.info( "creating %s from the following tables: %s" % (tablename, str(zip( tablenames, tracks ))))
    if min(tracks) != max(tracks):
        raise ValueError("number of rows not identical - will not create view" )

    from_statement = " , ".join( [ "%s as t%i" % (y[0],x) for x,y in enumerate(tables) ] )
    f = tables[0][1]
    where_statement = " AND ".join( ["t0.%s = t%i.%s" % (f,x+1,y[1]) for x,y in enumerate(tables[1:]) ] )
    
    all_columns, taken = [], set()
    for x, c in enumerate(columns):
        i = set(taken).intersection( set(c))
        if i:
            E.warn("duplicate column names: %s " % i )
            if not ignore_duplicates:
                table = tables[x][0]
                all_columns.extend( ["t%i.%s AS %s_%s" % (x,y,table,y) for y in i ] )
                c = [ y for y in c if y not in i ]
                
        all_columns.extend( ["t%i.%s" % (x,y) for y in c ] )
        taken.update( set(c) )

    all_columns = ",".join( all_columns)
    statement = '''CREATE %(view_type)s %(tablename)s AS SELECT t0.track, %(all_columns)s
                   FROM %(from_statement)s
                   WHERE %(where_statement)s
    ''' % locals()

    Database.executewait( dbhandle, statement )

    nrows = Database.executewait( dbhandle, "SELECT COUNT(*) FROM view_mapping" ).fetchone()[0]
    
    if nrows == 0:
        raise ValueError( "empty view mapping, check statement = %s" % (statement % locals()) )
    if nrows != min(tracks):
        E.warn( "view creates duplicate rows, got %i, expected %i" % (nrows, min(tracks)))

    E.info( "created view_mapping with %i rows" % nrows )

    touch( outfile )


def snip( filename, extension = None, alt_extension = None, path = False):
    '''return prefix of filename.

    If extension is given, make sure that filename has the extension.

    If path is set to false, the path is stripped from the file name.
    '''
    if extension: 
        if filename.endswith( extension ):
            root = filename[:-len(extension)]
        elif alt_extension and filename.endswith( alt_extension ):
            root = filename[:-len(alt_extension)]
        else:
            raise ValueError("'%s' expected to end in '%s'" % (filename, extension))
    else:
        root, ext = os.path.splitext( filename )

    if path==True: snipped = os.path.basename(root)
    else: snipped = root

    return snipped


def getCallerLocals(decorators=0):
    '''returns locals of caller using frame.

    optional pass number of decorators
    
    from http://pylab.blogspot.com/2009/02/python-accessing-caller-locals-from.html
    '''
    f = sys._getframe(2+decorators)
    args = inspect.getargvalues(f)
    return args[3]

def getCaller(decorators=0):
    frm = inspect.stack()[2+decorators]
    mod = inspect.getmodule(frm[0])
    return mod

def execute( statement, **kwargs ):
    '''execute a statement locally.'''

    if not kwargs: kwargs = getCallerLocals()

    kwargs = dict( PARAMS.items() + kwargs.items() )    

    L.debug("running %s" % (statement % kwargs))
    
    if "cwd" not in kwargs:
        cwd = os.getcwd()
    else:
        cwd = kwargs["cwd"]

    process = subprocess.Popen(  statement % kwargs,
                                 cwd = cwd, 
                                 shell = True,
                                 stdin = subprocess.PIPE,
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE )

    # process.stdin.close()
    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise PipelineError( "Child was terminated by signal %i: \nThe stderr was: \n%s\n%s\n" % (-process.returncode, stderr, statement ))

    return stdout, stderr

# Definition of helper functions for job scripts
# detect_pipe_error(): propagate error of programs not at the end of a pipe
# checkpoint(): exit a set of chained commands (via ;) if the previous command failed.
_exec_prefix = '''detect_pipe_error_helper() 
    {
    while [ "$#" != 0 ] ; do
        # there was an error in at least one program of the pipe
        if [ "$1" != 0 ] ; then return 1 ; fi
        shift 1
    done
    return 0 
    }
    detect_pipe_error() {
    detect_pipe_error_helper "${PIPESTATUS[@]}"
    return $?
    }
    checkpoint() {
        detect_pipe_error;
        if [ $? != 0 ]; then exit 1; fi;
    }
    '''

_exec_suffix = "; detect_pipe_error"

def buildStatement( **kwargs ):
    '''build statement from kwargs.

    Options in PARAMS are added, but kwargs take precedence.
    
    If outfile is in kwargs, 
    '''

    if "statement" not in kwargs:
        raise ValueError("'statement' not defined")

    local_params = substituteParameters( **kwargs )

    # build the statement
    try:
        statement = kwargs.get("statement") % local_params
    except KeyError, msg:
        raise KeyError( "Error when creating command: could not find %s in dictionaries" % msg)
    except ValueError, msg:
        raise ValueError( "Error when creating command: %s, statement = %s" % (msg, kwargs.get("statement") ) )

    # add bash as prefix to allow advanced shell syntax like 'wc -l <( gunzip < x.gz)'
    # executable option to call() does not work. Note that there will be an extra
    # indirection.
    statement = " ".join( re.sub( "\t+", " ", statement).split( "\n" ) ).strip()
    if statement.endswith(";"): statement = statement[:-1]

    L.debug( "running statement:\n%s" % statement )

    return statement

def expandStatement( statement ):
    '''add exec_prefix and exec_suffix to statement.'''
    
    return " ".join( (_exec_prefix, statement, _exec_suffix) )

def joinStatements( statements, infile ):
    '''join a chain of statements into a single statement.

    Each statement contains an @IN@ or a @OUT@ or both.
    These will be replaced by the names of successive temporary
    files.
    
    In the first statement, @IN@ is replaced with *infile*. 
    
    The last statement should move @IN@ to outfile.

    returns a single statement.
    '''
    
    prefix = getTempFilename()
    pattern = "%s_%%i" % prefix

    result = []
    for x, statement in enumerate(statements):
        if x == 0:
            s = re.sub( "@IN@", infile, statement )
        else:
            s = re.sub( "@IN@", pattern % x, statement )

        s = re.sub( "@OUT@", pattern % (x+1), s ).strip()

        if s.endswith(";"): s = s[:-1]
        result.append( s )

    assert prefix != ""
    result.append( "rm -f %s*" % prefix )
    
    result = "; checkpoint ; ".join( result )
    return result

def getStdoutStderr( stdout_path, stderr_path, tries=5 ):
    '''get stdout/stderr allowing for same lag.

    Try at most *tries* times. If unsuccessfull, throw PipelineError.

    Removes the files once they are read. 

    Returns tuple of stdout and stderr.
    '''
    x = tries
    while x >= 0:
        if os.path.exists( stdout_path ): break
        time.sleep(1)
        x -= 1
            
    x = tries
    while x >= 0:
        if os.path.exists( stderr_path ): break
        time.sleep(1)
        x -= 1

    try:
        stdout = open( stdout_path, "r" ).readlines()
    except IOError, msg:
        E.warn( "could not open stdout: %s" % msg )
        stdout = []
            
    try:
        stderr = open( stderr_path, "r" ).readlines()
    except IOError, msg:
        E.warn( "could not open stdout: %s" % msg )
        stderr = []

    try:
        os.unlink( stdout_path )
        os.unlink( stderr_path )
    except OSError, msg:
        pass

    return stdout, stderr

def run( **kwargs ):
    """run a statement.

    runs it on the cluster using drmaa if to_cluster is set.

    Troubleshooting:
       1. DRMAA creates sessions and their is a limited number
          of sessions available. If there are two many or sessions
          become not available after failed jobs, use ``qconf -secl``
          to list sessions and ``qconf -kec #`` to delete sessions.
       2. Memory: 1G of free memory can be requested using ``-l mem_free=1G``.
          If there are error messages like "no available queue", then the
          problem could be that a particular complex attribute has
          not been defined (the code should be ``hc`` for ``host:complex``
          and not ``hl`` for ``host:local``. Note that qrsh/qsub directly
          still works.
    """

    # combine options using correct preference
    options = dict(PARAMS.items())
    options.update( getCallerLocals().items() )
    options.update( kwargs.items() )

    def setupJob( session ):

        jt = session.createJobTemplate()
        jt.workingDirectory = os.getcwd()
        jt.jobEnvironment = { 'BASH_ENV' : '~/.bashrc' }
        jt.args = []
        jt.nativeSpecification = "-V -q %s -p %i -N %s %s" % \
            (options.get("job_queue", GLOBAL_OPTIONS.cluster_queue ),
             options.get("job_priority", GLOBAL_OPTIONS.cluster_priority ),
             "_" + re.sub( "[:]", "_", os.path.basename(options.get("outfile", "ruffus" ))),
             options.get("job_options", GLOBAL_OPTIONS.cluster_options))
            
        # keep stdout and stderr separate
        jt.joinFiles=False

        return jt

    shellfile = os.path.join( os.getcwd(), "shell.log" )

    pid = os.getpid()
    L.debug( 'task: pid = %i' % pid )

    # connect to global session 
    session = GLOBAL_SESSION
    L.debug( 'task: pid %i: sge session = %s' % (pid, str(session)))

    def buildJobScript( statement ):
        '''build job script from statement.

        returns (name_of_script, stdout_path, stderr_path)
        '''
        
        tmpfile = getTempFile( dir = os.getcwd() )
        tmpfile.write( "#!/bin/bash\n" ) #  -l -O expand_aliases\n" )
        tmpfile.write( 'echo "START--------------------------------" >> %s \n' % shellfile )
        # disabled - problems with quoting
        # tmpfile.write( '''echo 'statement=%s' >> %s\n''' % (shellquote(statement), shellfile) )
        tmpfile.write( "set &>> %s\n" % shellfile)
        tmpfile.write( "module list &>> %s\n" % shellfile )
        tmpfile.write( 'echo "END----------------------------------" >> %s \n' % shellfile )
        tmpfile.write( expandStatement( statement ) + "\n" )
        tmpfile.close()

        job_path = os.path.abspath( tmpfile.name )
        stdout_path = job_path + ".stdout" 
        stderr_path = job_path + ".stderr" 

        os.chmod( job_path, stat.S_IRWXG | stat.S_IRWXU )

        return (job_path, stdout_path, stderr_path)
    
    # run multiple jobs
    if options.get( "statements" ):

        statement_list = []
        for statement in options.get("statements"): 
            options["statement"] = statement
            statement_list.append(buildStatement( **options))
            
        if options.get( "dryrun", False ): return

        jt = setupJob( session )
        
        jobids, filenames = [], []
        for statement in statement_list:

            job_path, stdout_path, stderr_path = buildJobScript( statement )

            jt.remoteCommand = job_path
            jt.outputPath=":"+ stdout_path
            jt.errorPath=":" + stderr_path

            os.chmod( job_path, stat.S_IRWXG | stat.S_IRWXU )

            jobid = session.runJob(jt)
            jobids.append( jobid )
            filenames.append( (job_path, stdout_path, stderr_path) )

            L.debug( "job has been submitted with jobid %s" % str(jobid ))
        
        L.debug( "waiting for %i jobs to finish " % len(jobids) )
        session.synchronize(jobids, drmaa.Session.TIMEOUT_WAIT_FOREVER, False)
        
        # collect and clean up
        for jobid, statement, paths in zip( jobids, statement_list, filenames) :
            job_path, stdout_path, stderr_path = paths
            retval = session.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER)

            stdout, stderr = getStdoutStderr( stdout_path, stderr_path )

            if retval.exitStatus != 0:
                raise PipelineError( "---------------------------------------\n"
                                     "Child was terminated by signal %i: \n"
                                     "The stderr was: \n%s\n%s\n" 
                                     "---------------------------------------\n" % \
                                         (retval.exitStatus, 
                                          "".join( stderr),
                                          statement ) )

            try:
                os.unlink( job_path )
            except OSError:
                L.warn( "temporary job file %s not present for clean-up - ignored" % job_path )
            
        session.deleteJobTemplate(jt)

    # Run a single parallel job if
    #   1. job_queue is set, or
    #   2. to_cluster is not defined or to_cluster is set to True.
    # If the cluster has not been disabled through the command line, do not
    #     run on cluster
    elif (options.get( "job_queue" ) or 
          ("to_cluster" not in options or options.get( "to_cluster" ))) \
            and not GLOBAL_OPTIONS.without_cluster:

        statement = buildStatement( **options )

        if options.get( "dryrun", False ): return

        job_path, stdout_path, stderr_path = buildJobScript( statement )

        jt = setupJob( session )
        
        jt.remoteCommand = job_path
        # later: allow redirection of stdout and stderr to files; can even be across hosts?
        jt.outputPath=":"+ stdout_path
        jt.errorPath=":" + stderr_path

        if "job_array" in options and options["job_array"] != None:
            # run an array job
            start, end, increment = options.get("job_array" )
            L.debug("starting an array job: %i-%i,%i" % (start, end, increment ))
            # sge works with 1-based, closed intervals
            jobids = session.runBulkJobs( jt, start+1, end, increment )
            L.debug( "%i array jobs have been submitted as jobid %s" % (len(jobids), jobids[0]) )
            retval = session.synchronize(jobids, drmaa.Session.TIMEOUT_WAIT_FOREVER, True)
        else:
            jobid = session.runJob(jt)
            L.debug( "job has been submitted with jobid %s" % str(jobid ))
            try:
                retval = session.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER)
            except Exception, msg:
                # ignore message 24 in PBS
                # code 24: drmaa: Job finished but resource usage information and/or termination status could not be provided.":
                if not msg.message.startswith("code 24"): raise
                retval = None

        stdout, stderr = getStdoutStderr( stdout_path, stderr_path )

        if "job_array" not in options:
            if retval and retval.exitStatus != 0:
                raise PipelineError( "---------------------------------------\n"
                                     "Child was terminated by signal %i: \n"
                                     "The stderr was: \n%s\n%s\n"
                                     "-----------------------------------------" % \
                                         (retval.exitStatus, 
                                          "".join( stderr), statement))
            
        session.deleteJobTemplate(jt)
        try:
            os.unlink( job_path )
        except OSError:
            L.warn( "temporary job file %s not present for clean-up - ignored" % job_path )
    else:
        statement = buildStatement( **options )

        if options.get( "dryrun", False ): return
 
        if "<(" in statement:
            if "'" in statement: raise ValueError( "advanced bash syntax combined with single quotes" )
            statement = """/bin/bash -c '%s'""" % statement

        process = subprocess.Popen(  expandStatement( statement ),
                                     cwd = os.getcwd(), 
                                     shell = True,
                                     stdin = subprocess.PIPE,
                                     stdout = subprocess.PIPE,
                                     stderr = subprocess.PIPE )

        # process.stdin.close()
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            raise PipelineError( "---------------------------------------\n"
                                 "Child was terminated by signal %i: \n"
                                 "The stderr was: \n%s\n%s\n"
                                 "-----------------------------------------" % \
                                     (-process.returncode, stderr, statement ))

class MultiLineFormatter(logging.Formatter):
    '''logfile formatter: add identation for multi-line entries.'''

    def format(self, record):
        s = logging.Formatter.format(self, record)
        if record.message:
            header, footer = s.split(record.message)
            s = s.replace('\n', '\n' + ' '*len(header))
        return s

def submit( module, function, params = None,
            infiles = None, outfiles = None, 
            toCluster = True,
            logfile = None, 
            jobOptions = ""):
    '''Submit a python *function* as a job to the cluster.

    The function should reside in *module*. If *module* is
    not part of the PYTHONPATH, an absolute path can be given.

    *infiles* and *output* are either a single filename or a list of 
    input/output filenames. Neither options supports yet nested lists.
    '''

    if type( infiles ) in (list, tuple):
        infiles = " ".join( ["--input=%s" % x for x in infiles ] )
    else:
        infiles = "--input=%s" % infiles

    if type( outfiles ) in (list, tuple):
        outfiles = " ".join( ["--output=%s" % x for x in outfiles ] )
    else:
        outfiles = "--output=%s" % outfiles

    if logfile:
        logfile = "--log=%s" % logfile
    else:
        logfile = ""

    if params:
        params = "--params=%s" % ",".join(params)
    else:
        params = ""

    job_options = jobOptions 

    to_cluster = toCluster

    statement = '''python %(scriptsdir)s/run_function.py
                          --module=%(module)s
                          --function=%(function)s
                          %(logfile)s
                          %(infiles)s
                          %(outfiles)s
                          %(params)s
                '''
    run()

def clonePipeline( srcdir ):
    '''clone a pipeline from srcdir into the current directory.
    '''
    
    destdir = os.path.curdir

    copy_files = ("sphinxreport.ini", "conf.py", "pipeline.ini", "csvdb" )
    ignore_prefix = ("report", "_cache", "export", "tmp", "ctmp", "_static", "_templates" )

    def _ignore( p ):
        for x in ignore_prefix:
            if p.startswith( x ): 
                return True
        return False

    for root, dirs, files in os.walk(srcdir):

        relpath = os.path.relpath( root, srcdir )
        if _ignore( relpath ): continue

        for d in dirs:
            if _ignore( d ): continue
            dest = os.path.join( os.path.join(destdir, relpath, d ) )
            os.mkdir( dest )
            # touch
            s = os.stat( os.path.join(root, d ) )
            os.utime( dest, (s.st_atime, s.st_mtime ))

        for f in files:
            if _ignore( f ): continue

            fn = os.path.join( root, f )
            dest_fn = os.path.join( destdir, relpath, f ) 
            if f in copy_files:
                shutil.copyfile( fn, dest_fn )
            else:
                # realpath resolves links - thus links will be linked to
                # the original target
                os.symlink( os.path.realpath( fn),
                            dest_fn )

def writeConfigFiles( path ):
    
    for dest in ( "pipeline.ini", "sphinxreport.ini", "conf.py" ):
        src = os.path.join( path, dest)
        if os.path.exists(dest):
            L.warn( "file `%s` already exists - skipped" % dest )
            continue

        if not os.path.exists( src ):
            raise ValueError( "default config file `%s` not found"  % src )
        shutil.copyfile( src, dest )
        L.info( "created new configuration file `%s` " % dest )



def clean( patterns, dry_run = False ):
    '''clean up files given by glob *patterns*.

    returns list of files deleted together with their statinfo.
    '''

    cleaned = []

    for p in patterns:
        files = glob.glob( p )
        for x in files:
            statinfo = os.stat( x )
            cleaned.append( (x, statinfo) )
            if dry_run: continue
            os.unlink( x )
        L.info( "%i files: %s" % (len(files), p ))

    return cleaned

def peekParameters( workingdir, pipeline ):
    '''peak configuration parameters from an external directory.
    '''
    
    # attempt to locate directory with pipeline source code
    # This is a patch as pipelines might be called within the repository directory
    # or from an installed location
    dirname = os.path.dirname( __file__ )

    # called without a directory, use current directory
    if dirname == "":
        dirname = os.path.abspath(".")
    else:
        # else: use location of Pipeline.py
        # remove CGAT part, add CGATPipelines
        dirname = os.path.join( os.path.dirname(dirname), "CGATPipelines")
        # if not exists, assume we want version located
        # in directory of calling script.
        if not os.path.exists( dirname ):
            # directory is path of calling script
            v = getCallerLocals()
            dirname = os.path.dirname(v['__file__'])

    pipeline = os.path.join( dirname, pipeline )
    assert os.path.exists( pipeline ), "can't find pipeline source %s" % ( dirname, pipeline )
    if workingdir == "": workingdir = os.path.abspath(".")

    assert os.path.exists( workingdir ), "can't find working dir %s" % workingdir

    statement = "python %s -f -v 0 dump" % pipeline
    process = subprocess.Popen(  statement,
                                 cwd = workingdir, 
                                 shell = True,
                                 stdin = subprocess.PIPE,
                                 stdout = subprocess.PIPE,
                                 stderr = subprocess.PIPE )

    # process.stdin.close()
    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise PipelineError( "Child was terminated by signal %i: \nThe stderr was: \n%s\n" % (-process.returncode, stderr ))

    for line in stdout.split("\n"):
        if line.startswith("dump"):
            exec( line )

    return dump

def run_report( clean = True):
    '''run sphinxreport.'''

    dirname, basename = os.path.split( getCaller().__file__ )

    docdir = os.path.join( dirname, "pipeline_docs", snip( basename, ".py" ) )
    themedir = os.path.join( dirname, "pipeline_docs", "themes")
    relpath = os.path.relpath( docdir )
    trackerdir = os.path.join( docdir, "trackers" )

    # Requires an X11 connection to cluster nodes
    # A solution is to run xvfb on the nodes.
    to_cluster = False
    
    job_options= "-pe dedicated %i -R y" % PARAMS["report_threads"]

    # use a fake X display in order to avoid windows popping up
    # from R plots.
    xvfb_command = which("xvfb-run" )

    # permit multiple servers using -a option
    if xvfb_command: xvfb_command += " -a "
    else: xvfb_command = ""

    # if there is no DISPLAY variable set, xvfb runs, but
    # exits with error when killing process. Thus, ignore return
    # value.
    print os.getenv("DISPLAY"), "command=", xvfb_command
    if not os.getenv("DISPLAY"):
        erase_return = "|| true"
    else:
        erase_return = ""

    if clean: clean = """rm -rf report _cache _static;"""
    else: clean = ""

    statement = '''
    %(clean)s
    ( export SPHINX_DOCSDIR=%(docdir)s; 
      export SPHINX_THEMEDIR=%(themedir)s; 
    %(xvfb_command)s
    sphinxreport-build 
           --num-jobs=%(report_threads)s
           sphinx-build 
                    -b html 
                    -d %(report_doctrees)s
                    -c . 
           %(docdir)s %(report_html)s
    >& report.log %(erase_return)s )
    '''

    run()


USAGE = '''
usage: %prog [OPTIONS] [CMD] [target]

Execute pipeline %prog.

Commands can be any of the following

make <target>
   run all tasks required to build *target*

show <target>
   show tasks required to build *target* without executing them

plot <target>
   plot image (using inkscape) of pipeline state for *target*

config
   write new configuration files pipeline.ini, sphinxreport.ini and conf.py
   with default values

dump
   write pipeline configuration to stdout

touch
   touch files only, do not run

clone <source>
   create a clone of a pipeline in <source> in the current directory. The cloning
   process aims to use soft linking to files (not directories) as much as possible. 
   Time stamps are preserved. Cloning is useful if a pipeline needs to be re-run
   from a certain point but the original pipeline should be preserved.

'''

def main( args = sys.argv ):

    global GLOBAL_OPTIONS
    global GLOBAL_ARGS
    global GLOBAL_SESSION

    parser = optparse.OptionParser( version = "%prog version: $Id: Pipeline.py 2799 2009-10-22 13:40:13Z andreas $",
                                    usage = USAGE )
    
    parser.add_option( "--pipeline-action", dest="pipeline_action", type="choice",
                       choices=("make", "show", "plot", "dump", "config", "clone" ),
                       help="action to take [default=%default]." )

    parser.add_option( "--pipeline-format", dest="pipeline_format", type="choice",
                      choices=( "dot", "jpg", "svg", "ps", "png" ),
                      help="pipeline format [default=%default]." )

    parser.add_option( "-n", "--dry-run", dest="dry_run", action="store_true",
                      help="perform a dry run (do not execute any shell commands) [default=%default]." )

    parser.add_option( "-f", "--force", dest="force", action="store_true",
                      help="force running the pipeline even if there are uncommited changes "
                           "in the repository [default=%default]." )

    parser.add_option( "-l", "--local", dest="without_cluster", action="store_true",
                      help="execute all jobs locally [default=%default]." )

    parser.add_option( "-p", "--multiprocess", dest="multiprocess", type="int",
                       help="number of parallel processes to use (different from number of jobs to use for cluster jobs) [default=%default]." ) 

    parser.add_option( "-t", "--tempdir", dest="tempdir", type="string",
                       help="temporary directory to use [default=%default].")

    parser.add_option( "-e", "--exceptions", dest="log_exceptions", action="store_true",
                      help="echo exceptions immediately as they occur [default=%default]." )

    parser.add_option( "-i", "--terminate", dest="terminate", action="store_true",
                      help="terminate immediately at the first exception [default=%default]." )

    parser.set_defaults(
        pipeline_action = None,
        pipeline_format = "svg",
        pipeline_targets = [],
        multiprocess = 2,
        logfile = "pipeline.log",
        dry_run = False,
        without_cluster = False,
        force = False,
        log_exceptions = False,
        exceptions_terminate_immediately = False,
        )

    (options, args) = E.Start( parser, 
                               add_cluster_options = True )

    GLOBAL_OPTIONS, GLOBAL_ARGS = options, args
    PARAMS["dryrun"] = options.dry_run

    version = None

    try:
        # this is for backwards compatibility
        # get mercurial version
        repo = hgapi.Repo( PARAMS["scriptsdir"] )
        version = repo.hg_id()

        status = repo.hg_status()
        if status["M"] or status["A"]:
            if not options.force:
                raise ValueError( "uncommitted change in code repository at '%s'. Either commit or use --force" % PARAMS["scriptsdir"])
            else:
                E.warn( "uncommitted changes in code repository - ignored ")
        version = version[:-1]
    except:
        # try git:
        try:
            stdout, stderr = execute( "git rev-parse HEAD", cwd = PARAMS["scriptsdir"] )
        except: 
            stdout = "NA"
        version = stdout

    if args: 
        options.pipeline_action = args[0]
        if len(args) > 1:
            options.pipeline_targets.extend( args[1:] )

    if options.pipeline_action in ("make", "show", "svg", "plot", "touch" ):

        try:
            if options.pipeline_action == "make":

                # create the session proxy
                GLOBAL_SESSION = drmaa.Session()
                GLOBAL_SESSION.initialize()
                
                #
                #   make sure we are not logging at the same time in different processes
                #
                #session_mutex = manager.Lock()

                # set up extra file logger
                handler = logging.FileHandler( filename = options.logfile, 
                                               mode = "a" )
                handler.setFormatter( MultiLineFormatter( '%(asctime)s %(levelname)s %(module)s.%(funcName)s.%(lineno)d %(message)s' ) )
                logger = logging.getLogger()
                logger.addHandler( handler )
                
                L.info( E.GetHeader() )
                L.info( "code location: %s" % PARAMS["scriptsdir"] )
                L.info( "code version: %s" % version )

                pipeline_run( options.pipeline_targets, 
                              multiprocess = options.multiprocess, 
                              logger = logger,
                              verbose = options.loglevel,
                              log_exceptions = options.log_exceptions,
                              exceptions_terminate_immediately = options.exceptions_terminate_immediately,
                              )

                L.info( E.GetFooter() )

                GLOBAL_SESSION.exit()

            elif options.pipeline_action == "show":
                pipeline_printout( options.stdout, options.pipeline_targets, verbose = options.loglevel )

            elif options.pipeline_action == "touch":
                pipeline_run( options.pipeline_targets, 
                              touch_files_only = True,
                              verbose = options.loglevel )

            elif options.pipeline_action == "svg":
                pipeline_printout_graph( options.stdout, 
                                         options.pipeline_format,
                                         options.pipeline_targets )

            elif options.pipeline_action == "plot":
                outf, filename = tempfile.mkstemp()
                pipeline_printout_graph( os.fdopen(outf,"w"),
                                         options.pipeline_format,
                                         options.pipeline_targets )
                execute( "inkscape %s" % filename ) 
                os.unlink( filename )

        except ruffus_exceptions.RethrownJobError, value:
            E.error("some tasks resulted in errors - error messages follow below" )
            # print value
            E.error( value )
            E.error( "end of error messages" )
            raise

    elif options.pipeline_action == "dump":
        # convert to normal dictionary (not defaultdict) for parsing purposes
        print "dump = %s" % str(dict(PARAMS))

    elif options.pipeline_action == "config":
        f = sys._getframe(1)
        caller = inspect.getargvalues(f).locals["__file__"]
        prefix = os.path.splitext(caller)[0]
        writeConfigFiles( prefix )

    elif options.pipeline_action == "clone":
        clonePipeline( options.pipeline_targets[0] )

    else:
        raise ValueError("unknown pipeline action %s" % options.pipeline_action )

    E.Stop()

if __name__ == "__main__":

    main()

    #parser = optparse.OptionParser( version = "%prog version: $Id: Pipeline.py 2799 2009-10-22 13:40:13Z andreas $")

    #(options, args) = E.Start( parser, add_cluster_options = True )

    #global GLOBAL_OPTIONS
    #global GLOBAL_ARGS
    #GLOBAL_OPTIONS, GLOBAL_ARGS = options, args

    #L.info("in main")

    #run( **{"statement" : "printenv > test.out", "job_queue" : "server_jobs.q", "job_priority" : -10 } )
    #run( **{"statement" : "printenv > test2.out", "job_queue" : "server_jobs.q", "job_priority" : -10 } )
    #run( **{"statement" : "printenv > test3.out", "job_queue" : "server_jobs.q", "job_priority" : -10 } )
    



from pysam.csamtools cimport *

import collections, array, struct, itertools
import CGAT.Experiment as E

def merge_pairs( Samfile input_samfile,
                 outfile,
                 min_insert_size = 0,
                 max_insert_size = 400,
                 bed_format = None ):
    '''merge paired ended data.

    For speed reasons, the aligned region is only approximated using
    the read length. Indels within reads are ignored. Thus bed coordinates
    might be a few residues off.

    The strand is always set to '+'.

    Pairs with a maximum insert size larger than *max_insert_size* are removed.

    If `bed_format` is a number, only the first x columns will be output.
    '''

    cdef int ninput = 0
    cdef int nremoved_unmapped = 0
    cdef int nremoved_insert = 0
    cdef int nremoved_contig = 0
    cdef int nremoved_unpaired = 0
    cdef int noutput = 0
    cdef int flag

    cdef AlignedRead read
    cdef int c_max_insert_size = max_insert_size
    cdef int c_min_insert_size = min_insert_size
    cdef int start, end
    cdef int take_columns = 6
    
    if bed_format != None:
        if bed_format < 3 or bed_format > 6: 
            raise ValueError("a bed file must have at least 3 and at most 6 columns")
        take_columns = bed_format

    for read in input_samfile:

        ninput += 1

        flag = read._delegate.core.flag 
        # remove unmapped reads
        if flag & 4:
            nremoved_unmapped += 1
            continue

        # remove unpaired
        if not flag & 2:
            nremoved_unpaired += 1
            continue
            
        # this is second pair of read - skip
        if flag & 128:
            continue

        if read.tid != read.mrnm:
            nremoved_contig += 1
            continue

        if (c_max_insert_size and read.isize > c_max_insert_size) or (c_min_insert_size and read.isize < c_min_insert_size) :
            nremoved_insert += 1
            continue

        if read.pos < read.mpos:
            start = read.pos
            end = read.mpos + read.rlen
        else:
            start = read.mpos
            end = read.pos + read.rlen

        # count output pair as two so that it squares with ninput
        noutput += 2

        if take_columns == 3:
            outfile.write( "%s\t%i\t%i\n" %
                           (input_samfile.getrname( read.tid ),
                            start, end))
        elif take_columns == 4:
            outfile.write( "%s\t%i\t%i\t%s\n" % 
                           (input_samfile.getrname( read.tid ),
                            start, end,
                            read.qname))
        elif take_columns == 5:
            outfile.write( "%s\t%i\t%i\t%s\t%i\n" % 
                           (input_samfile.getrname( read.tid ),
                            start, end,
                            read.qname,
                            read.mapq,
                            ) )
        else:
            outfile.write( "%s\t%i\t%i\t%s\t%i\t%s\n" % 
                           (input_samfile.getrname( read.tid ),
                            start, end,
                            read.qname,
                            read.mapq,
                            "+"
                            ) )

    c = E.Counter()
    c.input = ninput
    c.removed_insert = nremoved_insert
    c.removed_contig = nremoved_contig
    c.removed_unmapped = nremoved_unmapped
    c.removed_unpaired = nremoved_unpaired
    c.output = noutput

    return c



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
WrapperZinba.py - wrap zinba peak caller
========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Runs the zinba peak caller.

Note that for zinba to work, a mappability file needs to have
been created.

The output bed files contain the P-value as their score field.

Usage
-----

Documentation
-------------

Code
----

'''

import os, sys, re, optparse, tempfile, shutil, subprocess
import collections

import Experiment as E
import IOTools

## for zinba
from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.vectors as rovectors
from rpy2.rinterface import RRuntimeError

def bamToBed( infile, outfile ):
    '''convert bam to bed with bedtools.'''

    statement = "bamToBed -i %(infile)s > %(outfile)s" % locals()

    E.debug( "executing statement '%s'" % statement )

    retcode = subprocess.call(  statement,
                                cwd = os.getcwd(), 
                                shell = True )
    if retcode < 0:
        raise OSError( "Child was terminated by signal %i: \n%s\n" % (-retcode, statement ))

    return outfile

ZinbaPeak = collections.namedtuple( "ZinbaPeak", "contig unrefined_start unrefined_end strand posterior summit height refined_start refined_end median fdr" )

def iteratePeaks( infile ):
    '''iterate of zinba peaks in infile.'''
    
    for line in infile:

        if line.startswith("#"): continue
        if line.startswith("PEAKID\tChrom"): continue
        # skip empty lines
        if line.startswith("\n"): continue

        data = line[:-1].split("\t")

        if len(data) != 12:
            raise ValueError( "could not parse line %s" % line )

        # I assume these are 1-based coordinates
        data[2] = max(int(data[2]) - 1, 0)
        # end
        data[3] = int(data[3])
        # posterior
        data[5] = float(data[5])
        # summit
        data[6] = max(int(data[6]) - 1, 0)
        # height
        data[7] = int(data[7])
        # refined_start
        data[8] = max(int(data[8]) - 1, 0)
        # end
        data[9] = int(data[9])
        # median
        data[10] = int(data[10])
        # qvalue
        data[11] = float(data[11])

        yield ZinbaPeak._make( data[1:] )

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-f", "--input-format", dest="input_format", type="choice",
                      choices = ("bed", "bam"),
                      help="input file format [default=%default]."  )
    
    parser.add_option("-s", "--fragment-size", dest="fragment_size", type="int",
                      help="fragment size [default=%default]."  )

    parser.add_option("-m", "--mappability-dir", dest="mappability_dir", type="string",
                      help="mappability_dir [default=%default]."  )

    parser.add_option("-b", "--bit-filename", dest="bit_filename", type="string",
                      help="2bit genome filename [default=%default]."  )

    parser.add_option("-c", "--control-filename", dest="control_filename", type="string",
                      help="filename of input/control data in bed format [default=%default]."  )

    parser.add_option("-i", "--index-dir", dest="index_dir", type="string",
                      help="index directory [default=%default]."  )

    parser.add_option("-t", "--threads", dest="threads", type="int",
                      help="number of threads to use [default=%default]."  )

    parser.add_option("-q", "--fdr-threshold", dest="fdr_threshold", type="float",
                      help="fdr threshold [default=%default]."  )

    parser.add_option("-a", "--alignability-threshold", dest="alignability_threshold", type="int",
                      help="alignability threshold [default=%default]."  )

    parser.add_option("-p", "--per-contig", dest="per_contig", action = "store_true",
                      help="run analysis per chromosome [default=%default]")

    parser.add_option("-w", "--temp-dir", dest="tempdir", type="string",
                      help="use existing directory as temporary directory [default=%default]."  )

    parser.add_option( "--keep-temp", dest="keep_temp", action = "store_true",
                      help="keep temporary directory [default=%default]")

    parser.add_option( "--action", dest="action", type="choice",
                       choices=("full", "count", "predict", "model"),
                       help="action to perform [default=%default]")

    parser.add_option( "--improvement", dest="improvement", type="float",
                       help="relative improvement of likelihood until convergence [default=%default]")
    
    parser.set_defaults(
        input_format = "bed",
        fragment_size = 200,
        mappability_dir = None,
        threads = 1,
        alignability_threshold = 1,
        bit_filename = None,
        fdr_threshold = 0.05,
        tempdir = None,
        winsize = 250,
        offset = 125,
        cnvWinSize = 1e+05,
        cnvOffset = 2500,
        per_contig = False,
        keep_temp = False,
        filelist = "files.list",
        action = "full",
        improvement = 0.00001,
        )


    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if len(args) != 2:
        raise ValueError("please specify a filename with sample data and an output file")

    filename_sample, filename_output = args[0], args[1]
    filename_control = options.control_filename
    
    # load Zinba
    R.library( 'zinba' )

    if not options.tempdir:
        tmpdir = tempfile.mkdtemp( )
    else:
        tmpdir = options.tempdir

    E.debug( "temporary files are in %s" % tmpdir )

    if options.input_format == "bam":
        E.info( "converting bam files to bed" )
        if not os.path.exists( os.path.join( tmpdir, "sample.bed")):
            filename_sample = bamToBed( filename_sample, os.path.join( tmpdir, "sample.bed" ) )
        else:
            E.info("using existing file %(tmpdir)s/sample.bed" % locals() )
            filename_sample = os.path.join( tmpdir, "sample.bed")
        if filename_control:
            if not os.path.exists( os.path.join( tmpdir, "control.bed")):
                filename_control = bamToBed( filename_control, os.path.join( tmpdir, "control.bed" ) )
            else:
                E.info("using existing file %(tmpdir)s/control.bed" % locals() )
                filename_control = os.path.join( os.path.join( tmpdir, "control.bed"))

    fragment_size = options.fragment_size
    threads = options.threads
    bit_filename = options.bit_filename
    mappability_dir = options.mappability_dir
    fdr_threshold = options.fdr_threshold
    tol = options.improvement

    contigs = E.run( "twoBitInfo %(bit_filename)s %(tmpdir)s/contig_sizes" % locals() )
    contig2size = dict( [x.split() for x in IOTools.openFile( os.path.join( tmpdir, "contig_sizes")) ] )

    outdir = filename_output + "_files" 
    if not os.path.exists( outdir ):
        os.mkdir( outdir )
        
    filelist = os.path.join( outdir, filename_output + ".list")
    modelfile = os.path.join( outdir, filename_output + ".model")
    winfile = os.path.join( outdir, filename_output + ".wins")
    winSize=250
    offset=125
    cnvWinSize=100000
    cnvOffset=0
    winGap = 0
    peakconfidence = 1.0 - fdr_threshold

    if not os.path.exists( os.path.join( tmpdir, "basecount")):
        E.info( "computing counts" )

        R( '''basealigncount( inputfile='%(filename_sample)s',
                          outputfile='%(tmpdir)s/basecount',
                          extension=%(fragment_size)i,
                          filetype='bed',
                          twoBitFile='%(bit_filename)s' )
                          '''  % locals() )
    else:
        E.info( "using existing counts" )

    # tried incremental updates
    # for contig, size in contig2size.iteritems():
    #     for size in 
    #     fn = os.path.join( tmpdir, "sample_%(contig)s_win%(size)ibp_offset(offset)ibp.txt" % locals() )
    if options.action == "count":

        E.info("computing window counts only - saving results in %s" % outdir )
        R('''buildwindowdata(
                     seq='%(filename_sample)s', 
                     align='%(mappability_dir)s',
                     input='%(filename_control)s', 
                     twoBit='%(bit_filename)s', 
                     winSize=%(winSize)i,
                     offset=%(offset)i,
                     cnvWinSize=%(cnvWinSize)i,
                     cnvOffset=%(cnvOffset)i,
                     filelist='%(filelist)s',
                     filetype='bed',  
                     extension=%(fragment_size)s,
                     outdir='%(outdir)s/') ''' % locals() )

    elif options.action == "model":

        # The important option is buildwin = 0
        # parameterized for broad == FALSE and input present
        # see zinba.R
        # model selection only on chr19.
        R('''run.zinba( 
                filelist='%(filelist)s',
                formula=NULL,formulaE=NULL,formulaZ=NULL,
                outfile='%(filename_output)s',
                seq='%(filename_sample)s', 
                input='%(filename_control)s', 
                filetype='bed',  
                align='%(mappability_dir)s',
                twoBit='%(bit_filename)s', 
                extension=%(fragment_size)s, 
                winSize=%(winSize)i,
                offset=%(offset)i,
                cnvWinSize=%(cnvWinSize)i,
                cnvOffset=%(cnvOffset)i,
                basecountfile='%(tmpdir)s/basecount',
                buildwin=0,
                threshold=%(fdr_threshold)f,
                pquant=1,
                peakconfidence=%(peakconfidence)f,
                winGap=%(winGap)i,
                tol=%(tol)f,
                initmethod="count",
                method="mixture",
                numProc=%(threads)i,
                printFullOut=1,
                interaction=FALSE,
                selectmodel=TRUE,
                selectchr='chr19',
                selectcovs=c("input_count"),
                selecttype="complete",
                FDR=TRUE)''' % locals())

    elif options.action == "predict":
    
        # The important option is buildwin = 0 and selectmodel = FALSE
        # parameterized for broad == FALSE and input present
        # see zinba.R
        # model selection only on chr19.
        if not os.path.exists( modelfile ):
            raise OSError( "model file %s does not exist" )

        E.info( "reading model from %s" % modelfile )

        R('''
        final=read.table('%(modelfile)s', header=T, sep="\t")
        final=final[final$fail==0,]
        bestBIC=which.min(final$BIC)
        formula=as.formula(paste("exp_count~",final$formula[bestBIC]))
        formulaE=as.formula(paste("exp_count~",final$formulaE[bestBIC]))
        formulaZ=as.formula(paste("exp_count~",final$formulaZ[bestBIC]))
        cat("Background formula is:\n\t")
        print(formula)
        cat("Enrichment formula is:\n\t")
        print(formulaE)
        cat("Zero-inflated formula is:\n\t")
        print(formulaE)
        ''' % locals() )

        E.info( "predicting peaks" )

        R('''run.zinba(
                filelist='%(filelist)s',
                outfile='%(filename_output)s',
                seq='%(filename_sample)s',
                input='%(filename_control)s',
                filetype='bed',
                align='%(mappability_dir)s',
                twoBit='%(bit_filename)s',
                extension=%(fragment_size)s,
                winSize=%(winSize)i,
                offset=%(offset)i,
                cnvWinSize=%(cnvWinSize)i,
                cnvOffset=%(cnvOffset)i,
                basecountfile='%(tmpdir)s/basecount',
                buildwin=0,
                threshold=%(fdr_threshold)f,
                pquant=1,
                winGap=%(winGap)i,
                initmethod="count",
                tol=%(tol)f,
                method="mixture",
                numProc=%(threads)i,
                printFullOut=1,
                interaction=FALSE,
                selectmodel=FALSE,
                formula=formula,
                formulaE=formulaE,
                formulaZ=formulaZ,
                peakconfidence=%(peakconfidence)f,
                FDR=TRUE)''' % locals())

    elif options.action == "per_contig":

        E.info("processing per chromosome" )
        for contig, size in contig2size.iteritems():
            if contig not in ("chr16",): continue

            E.info("processing contig %s" % contig)
            filename_sample_contig = filename_sample + "_%s" % contig
            filename_control_contig = filename_control + "_%s" % contig
            if not os.path.exists( filename_output + "_files" ):
                os.mkdir( filename_output + "_files" )
            filename_output_contig = os.path.join( filename_output + "_files", contig )
            filename_basecounts_contig = os.path.join( tmpdir, "basecount_%s" % contig)

            E.run( "grep %(contig)s < %(filename_sample)s > %(filename_sample_contig)s" % locals() )
            E.run( "grep %(contig)s < %(filename_control)s > %(filename_control_contig)s" % locals() )

            if not os.path.exists( filename_basecounts_contig ):
                E.info( "computing counts" )

                R( '''basealigncount( inputfile='%(filename_sample_contig)s',
                                  outputfile='%(filename_basecounts_contig)s',
                                  extension=%(fragment_size)i,
                                  filetype='bed',
                                  twoBitFile='%(bit_filename)s' )
                                  '''  % locals() )
            else:
                E.info( "using existing counts" )

            # run zinba, do not build window data
            R( '''zinba( refinepeaks=1,
                     seq='%(filename_sample_contig)s',
                     input='%(filename_control_contig)s',
                     filetype='bed',
                     align='%(mappability_dir)s',
                     twoBit='%(bit_filename)s',
                     outfile='%(filename_output_contig)s',
                     extension=%(fragment_size)s,
                     basecountfile='%(filename_basecounts_contig)s',
                     numProc=%(threads)i,
                     threshold=%(fdr_threshold)f,
                     broad=FALSE,
                     printFullOut=0,
                     interaction=FALSE,
                     mode='peaks',
                     FDR=TRUE) '''  % locals() )
    elif options.action == "full":

        # run zinba, do not build window data
        R( '''zinba( refinepeaks=1,
                     seq='%(filename_sample)s',
                     input='%(filename_control)s',
                     filetype='bed',
                     align='%(mappability_dir)s',
                     twoBit='%(bit_filename)s',
                     outfile='%(filename_output)s',
                     extension=%(fragment_size)s,
                     basecountfile='%(tmpdir)s/basecount',
                     numProc=%(threads)i,
                     threshold=%(fdr_threshold)f,
                     broad=FALSE,
                     printFullOut=0,
                     interaction=FALSE,
                     mode='peaks',
                     FDR=TRUE) '''  % locals() )

    if not (options.tempdir or options.keep_temp):
        shutil.rmtree( tmpdir )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
filter_tokens.py - select columns from a table 
==============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python filter_tokens.py --help

Type::

   python filter_tokens.py --help

for command line help.

Documentation
-------------

Code
----

'''
import sys
import re
import string
import os
import time
import optparse

USAGE = """python %s [token1 [token2 [...]]] < stdin > stdout

filter lines where pattern matches

""" % sys.argv[0]

import CGAT.Experiment as E


if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: filter_tokens.py 2782 2009-09-10 11:40:29Z andreas $", usage = globals()["__doc__"])

    parser.add_option( "-r", "--regex-token", dest="regex_token", type="string",
                       help="regular expression for tokens (has to create one pair of brackets)." )

    parser.add_option( "-f", "-a", "--apply", "--filename-tokens", dest="filename_tokens", type="string",
                       help="filename with tokens." )

    parser.add_option( "-o", "--tokens", "--tokens", dest="tokens", type="string",
                       help="',' separated list of tokens [default=%default]." )

    parser.add_option( "-c", "--columns", dest="columns", type="string",
                       help="comma separated list of columns." )
    
    parser.add_option( "-i", "--invert-match", dest="invert_match", action="store_true",
                       help="only keep non-matching entries.")

    parser.add_option( "-t", "--no-titles", dest="titles", action="store_false",
                       help="no titles.")

    parser.set_defaults(
        regex_token = "^(\S+)",
        filename_tokens= None,
        columns = None,
        invert = False,
        titles = True,
        tokens = None,
        )

    (options, args) = E.Start( parser )

    if options.columns:
        options.columns=map(lambda x: int(x) -1, options.columns.split(","))

    if len(args) == 0 and not options.filename_tokens:
        print USAGE, "please specify tokens on command line or supply file with tokens."
        sys.exit(1)

    regex_token = re.compile(options.regex_token)

    file_id = 0

    ##########################################
    ## build list of keys to keep
    keys = {}

    for a in args: keys[a] = 1
    
    if options.filename_tokens:
        infile = open(options.filename_tokens, "r")
        for line in infile:
            if line[0] == "#": continue
            a = line[:-1].split("\t")[0]
            keys[a] = 1
            
    if options.tokens:
        for a in options.tokens.split( "," ):
            keys[a] = 1
        
    ##########################################

    ninput = 0
    nkept = 0

    for line in sys.stdin:

        if line[0] == "#": continue
        if line[0] == ">":
            print line[:-1]
            continue

        ninput += 1

        ## skip first line if there are titles
        if ninput == 1 and options.titles:
            print line[:-1]
            continue
        
        found = False

        if options.columns:
            data = line[:-1].split("\t")
            for c in options.columns:
                try:
                    if data[c] in keys:
                        found = True
                        break
                except IndexError:
                    continue

        elif regex_token:
            r = regex_token.search( line[:-1] )
            if r:
                if keys.has_key( r.groups()[0]):
                    found = True

        if options.invert_match:
            found = not found
            
        if found:
            nkept += 1
            print line[:-1]

    if options.loglevel >= 1:
        options.stdlog.write( "# input=%i, kept=%i, discarded=%i\n" % (ninput, nkept, ninput-nkept) )
        
    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
align_mali_vs_mali.py - align two multiple alignments
=====================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script does a profile vs profile alignment of two
multiple alignments.

This script is only parameterized for the alignment of
protein sequences, but could be extended to the alignment
of DNA sequences as well.

Usage
-----

Example::

   python align_mali_vs_mali.py mali1.fasta mali2.fasta > mali_out.fasta

Type::

   python align_mali_vs_mali.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse

import CGAT.Experiment as E
import CGAT.Mali as Mali
import CGAT.IOTools as IOTools
import alignlib

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-o", "--gop", dest="gop", type="float",
                      help="gap opening penalty [default=%default]."  )

    parser.add_option("-e", "--gep", dest="gep", type="float",
                      help="gap extension penalty [default=%default]."  )

    parser.add_option("-m", "--mode", dest="mode", type="choice",
                      choices = ("global", "local" ),
                      help="alignment mode, global=nw, local=sw [default=%default]."  )

    parser.set_defaults(
        gop = -12.0,
        gep = -2.0,
        format= "fasta",
        mode = "local",
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if len(args) != 2: raise ValueError("please supply two multiple alignments in FASTA format.")

    mali1 = Mali.Mali()
    mali2 = Mali.Mali()

    E.info( "read 2 multiple alignments" )

    mali1.readFromFile( IOTools.openFile( args[0], "r" ), format=options.format )
    mali2.readFromFile( IOTools.openFile( args[1], "r" ), format=options.format )

    cmali1 = Mali.convertMali2Alignlib( mali1 )
    cmali2 = Mali.convertMali2Alignlib( mali2 )

    if options.mode == "local":
        mode = alignlib.ALIGNMENT_LOCAL
    elif options.mode == "global":
        mode = alignlib.ALIGNMENT_GLOBAL
        
    alignator = alignlib.makeAlignatorDPFull( mode,
                                              options.gop, options.gep )

    alignlib.setDefaultEncoder( alignlib.getEncoder( alignlib.Protein20) )
    alignlib.setDefaultLogOddor( alignlib.makeLogOddorDirichlet( 0.3 ) )
    alignlib.setDefaultRegularizor( alignlib.makeRegularizorDirichletPrecomputed() )

    cprofile1 = alignlib.makeProfile( cmali1 )
    cprofile2 = alignlib.makeProfile( cmali2 )

    result = alignlib.makeAlignmentVector()

    alignator.align( result, cprofile1, cprofile2 )

    E.debug( "result=\n%s" % alignlib.AlignmentFormatEmissions( result) )

    cmali1.add( cmali2, result )

    outmali = Mali.convertAlignlib2Mali( cmali1,
                                         identifiers = mali1.getIdentifiers() + mali2.getIdentifiers() )
    
    outmali.writeToFile( options.stdout, format=options.format)

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali_remove_gaps.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python mali_remove_gaps.py --help

Type::

   python mali_remove_gaps.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import math

USAGE="""python %s [OPTIONS] < exonerate_output > filtered

Prune a nucelotide multiple alignment according to a master sequence.

1. Go in codon steps through the multiple alignment according
to the master sequence.

2. Remove all columns in other sequences, that
        1. fall out of frame
        2. are incomplete codons

Version = $Id: mali_remove_gaps.py 2782 2009-09-10 11:40:29Z andreas $

Options:
-h, --help                      print this message.
-v, --verbose=                  loglevel.
-o, --file-output               output
""" % sys.argv[0]

param_long_options=["verbose=", "help", "file-output=", "version" ]

param_short_options="v:hm:e:p:c"

param_loglevel = 1

param_gap_char = "-"
param_mask_char = "x"

param_filename_output = None

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.MaliIO as MaliIO
import CGAT.Exons as Exons

##------------------------------------------------------------
if __name__ == '__main__':

    try:
        optlist, args = getopt.getopt(sys.argv[1:], param_short_options, param_long_options)
    except getopt.error, msg:
        print USAGE, msg
        sys.exit(2)

    for o,a in optlist:
        if o in ( "-v", "--verbose" ):
            param_loglevel = int(a)
        elif o in ( "--version", ):
            print "version="
            sys.exit(0)
        elif o in ( "-h", "--help" ):
            print USAGE
            sys.exit(0)
        elif o in ("-o", "--file-output"):
            param_filename_output = a
            
    ## 1. read multiple alignment in fasta format
    mali, identifiers = MaliIO.readFasta( sys.stdin )

    if param_loglevel >= 1:
        print "# read mali with %i entries." % len(identifiers)

    print E.GetHeader()
    print E.GetParams()

    ## 1. remove gaps in multiple alignment

    mali = MaliIO.removeGaps( mali )
    
    if param_master:
        frame_columns = GetFrameColumns( mali, param_master )
    elif param_master_pattern:
        columns = []
        for id in identifiers:
            if re.search( param_master_pattern, id ):
                columns += GetFrameColumns( mali, id )

        if len(columns) == 0:
            columns += GetFrameColumns( mali, identifiers[0] )
                
        # sort all columns by tuple. The "shortest" codon will be first (1,2,3) before (1,2,100)
        columns.sort()

        # select codons
        frame_columns = []
        last_codon = columns[0]
        for codon in columns[1:]:
            # skip identical codons
            if codon == last_codon: continue

            # take first (shortest) codon in case of identical first residue
            if codon[0] == last_codon[0]: continue
                
            # if not overlapping, keep
            if codon[0] > last_codon[2]:
                frame_columns.append( last_codon )

            # if overlapping, but out of register: skip
            last_codon = codon
            
        frame_columns.append( last_codon )        

    ## translate characters to upper/lower case according to exon info.
    if exons:
        for id in mali:
            if id in exons:
                mali[id] = AddExonInformation( mali[id], exons[id], mask_char = param_mask_char )

    if param_loglevel >= 1:
        print "# found %i columns" % (len(frame_columns))

    mask_chars = ( string.upper( param_mask_char ), string.lower( param_mask_char ) )

    
    for id in mali.keys():
        sequence = mali[id]
        fragments = []
        nstops, ncodons, naligned = 0, 0, 0
        for a,b,c in frame_columns:
            codon = sequence[a] + sequence[b] + sequence[c]

            codon_is_aligned = False
            codon_is_ok = True
            for x in codon:
                ## a codon will be masked, if it either
                ## 1. contains a gap character
                ## 2. is an unaligned character, i.e.,
                ##     exons and masked, or no exons and lowerwase
                residue_is_unaligned = (x == param_gap_char) or \
                                       (not exons and x in string.lowercase) or \
                                       (exons and x in mask_chars)
                codon_is_aligned = codon_is_aligned or not residue_is_unaligned
                codon_is_ok = codon_is_ok and not residue_is_unaligned

            if codon_is_aligned: naligned += 1
            
            if codon_is_ok:
                ncodons += 1
                if string.upper(codon) in ("TAG", "TAA", "TGA"):
                    if param_remove_stops:
                        fragments.append( param_gap_char * 3 )
                    else:
                        fragments.append( codon )                        
                    nstops += 1
                else:
                    fragments.append( codon )                                            
            else:
                fragments.append( param_gap_char * 3 )
                
        mali[id] = string.join(fragments, "")
        if param_loglevel >= 1:
            print "# sequence: %s\tpositions: %i\taligned:%i\tcodons: %i\t stops: %i" % (id, len(fragments), naligned, ncodons, nstops)
            sys.stdout.flush()
            
    for id in mali.keys():
        if param_mark_codons:
            a = mali[id]
            f = lambda x: a[x:x+3]
            s = string.join( [ f(x) for x in range(0,len(a),3)], " ")
        else:
            s = mali[id]
        print ">%s\n%s" % (id, s )

    if param_filename_translation:
        outfile = open( param_filename_translation, "w")
        for id in mali.keys():
            outfile.write( ">%s\n%s\n" % (id, Genomics.TranslateDNA2Protein(mali[id])))
        outfile.close()

    print E.GetFooter()
    
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
psql_add_tables.py - merge two tables
=====================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script will merge two tables by adding one to the other.
An index ``field`` is automatically incremented such that there
is no overlap between the two tables.

The column ``field`` will not necessary be continuous.

Usage
-----

Example::

   python psql_add_tables.py src dest field

will add table ``src`` to table ``dest`` using ``field``
as an incremental counter. The column ``field`` has to
present in both tables.

Type::

   python psql_add_tables.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import time
import sets
import optparse
import math
import tempfile
import copy

import pgdb

import CGAT.Experiment as E

USAGE="""python psql_add_tables.py src dest field"""


##---------------------------------------------------------------------------------------------
def DbExecute( dbhandle, statement ):

    cc = dbhandle.cursor()
    cc.execute(statement)
    result = cc.fetchall()
    cc.close()
    return result

##---------------------------------------------------------------------------------------------
def DbDo( dbhandle, statement ):

    cc = dbhandle.cursor()
    cc.execute(statement)
    dbhandle.commit()
    cc.close()

##------------------------------------------------------------------------------------------
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: psql_add_tables.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

    parser.add_option( "-s", "--start", dest="start", type="int",
                       help="index to start.")

    parser.add_option( "-f", "--fields", dest="fields", type="string",
                       action="append",
                       help="field with index.")

    parser.set_defaults( \
        start = None,
        fields = [] )

    (options, args) = E.Start( parser, add_psql_options = True )
    
    dbhandle = pgdb.connect( options.psql_connection )

    if len(args) != 2:
        raise "please supply src and dest."
    
    src_table, dest_table = args

    if len(options.fields) == 0:
        raise "please supply a field."

    statement = "SELECT %s FROM %s" % (options.fields[0], dest_table )

    indices = map(lambda x: int(x[0]), DbExecute( dbhandle, statement ))
    
    indices.sort()
    max_index = indices[-1]

    if options.start:
        if options.start < max_index:
            raise "start %i is less than maximum index %i" % (options.start, max_index)
        increment = options.start
    else:
        increment = max_index

    for field in options.fields:
        statement = "UPDATE %s SET %s=CAST(%s AS INT)+%i" % (src_table, field, field, increment )
        options.stdlog.write("# incrementing field %s in %s by %i\n" % (field, src_table, increment ) )
        DbDo( dbhandle, statement )

    statement = "INSERT INTO %s SELECT * FROM %s" % (dest_table, src_table)
    options.stdlog.write("# adding table %s to %s\n" % (src_table, dest_table) )
    DbDo( dbhandle, statement )
    
    E.Stop()

    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
split_links.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python split_links.py --help

Type::

   python split_links.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import time
import math
import CGAT.Experiment as E
import CGAT.WrapperAdaptiveCAI as WrapperAdaptiveCAI
import numpy

open_files = {}

def WriteLine( a, b, line, prefix="%s-%s" ):

    key1 = prefix % (a,b)
    key2 = prefix % (b,a)    
    if key1 in open_files or os.path.exists( key1 ):
        key = key1
    else:
        key = key2
    
    if key not in open_files:
        open_files[key] = open( key, "a" )

    f = open_files[key]
    f.write( line )
    f.flush()
    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: split_links.py 2781 2009-09-10 11:33:14Z andreas $")

    parser.add_option("-m", "--method", dest="method", type="string",
                      help="method for splitting." )

    parser.add_option("-r", "--regex", dest="regex", type="string",
                      help="regex to find prefix." )

    parser.add_option("-o", "--output", dest="output", type="string",
                      help="output filename." )

    parser.add_option("-t", "--targets", dest="targets", type="string",
                      help="output filename." )

    parser.set_defaults()

    (options, args) = E.Start( parser )

    if options.targets: options.targets = options.targets.split(",")

    nsame = 0
    ndiff = 0
    
    if options.method=="prefix":

        for line in sys.stdin:
            if line[0] == "#": continue

            data = line[:-1].split("\t")

            g1 = re.search( options.regex, data[0] ).groups()[0]
            g2 = re.search( options.regex, data[1] ).groups()[0]            

            if g1 == g2:
                for t in options.targets:
                    if g1 == t: continue
                    WriteLine( g1, t, line, options.output )
                    nsame += 1
            else:
                WriteLine( g1, g2, line, options.output )
                ndiff += 1

    print "nsame=%i, ndiff=%i" % (nsame, ndiff )

    E.Stop()
            
        
    
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2mali.py - operate on a multiple alignments
===============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

perform operations (masking, renaming, conversion, ...) on a multiple alignment

Available edit operations are:

remove-unaligned-ends
   removes unaligned ends (lowercase characters)

remove-end-gaps
   removes gapped ends.

shift-alignment
   add offset to locations. This requires one parameter, which
   should be filename of a file containing id-offset in tabular form.

mark-transitions
   marks transitions in the alignment. The parameter denotes either

      1 a filename to a file containing id-transition pairs in
        tabular form. There can be multiple transitions for the same id, but
        they should be on separate lines. Use 'mali' as an id for multiple
        alignment coordinates.

      2 a colon (':') -separated list of transitions

filter-even-transitions
   same parameter as above, but will remove all even segments (0, 2, ...).

keep-odd-segments
   synonym to filter-even-transitions

filter-odd-transitions
   same parameter as above, but will remove all odd segments (1, 3, ...).

keep-even-segments
   synonym to filter-odd-transitions

lower
   convert the multiple alignment to lowercase characters.

upper
   convert the multiple alignment to uppercase characters.

remove-all-gaps
   remove columns that are completely gaps.

remove-any-gaps
   remove columns that contain at least one gap.

remove-some-gaps
   remove columns that contain some gaps. The amount is specified in
   the parameter list as minimum_gaps. A value of
   --parameter=3 will remove all columns with three or more gaps.

add-annotation
   add annotation to mali.

recount
   recount alignment ranges.

remove-empty-sequences
   removes all empty sequences from mali.

remove-unaligned-pairs
   removes unaligned pairs in the multiple alignment. 

filter-3rd
   return alignment with only third codon positions.

filter-4d               
   return alignment with only the four-fold degenerate site.

mask-random-proportion  
   mask random proportion of colums

exclude-with-stop
   remove all sequences that contain a stop codon

exclude-with-frameshift
   remove all sequences that contain an indel that is not
   a multiple of three. Indels are determined with respect to the first
   sequence.

Parameters are given to the option parameters in a comma-separated list in the order
that the edit operations are called upon.

Usage
-----

Example::

   python mali2mali.py --help

Type::

   python mali2mali.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math
import time

import CGAT.Genomics as Genomics
import CGAT.Masker as Masker
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Mali as Mali

##------------------------------------------------------------
##------------------------------------------------------------
##------------------------------------------------------------
def removeUnalignedPairs( mali, options ):
    """remove unaligned pairs in multiple alignment.

    This procedure works by finding the column that has most sequences
    aligned. Only keep sequences, that are aligned in this region.
    """
    ids = mali.getIdentifiers()

    nids = len(ids)
    columns = []
    max_size = 0

    ## compute columns with maximum occupancy
    for x in range(mali.getWidth()):
        aligned = set()
        for y in range(nids):
            if mali[ids[y]][x] not in options.gap_chars:
                aligned.add( y )
        if len(aligned) > max_size:
            columns = []
            max_size = len(aligned)
        if len(aligned) == max_size:
            columns.append( (x, aligned) )

    ## compute sequences contributing most to occupancy
    counts = {}
    for col, s in columns:
        ss = list(s)
        ss.sort()
        ss = tuple(ss)
        if ss not in counts: counts[ss] = 0
        counts[ss] += 1

    ## compute sequences contributing most to occupancy
    cc = []
    for s, c in counts.items(): cc.append( (c, s) )
    cc.sort()

    take_rows = set(cc[-1][1])

    delete_rows = [ ids[x] for x in set(range(len(ids))).difference( set(take_rows) )]
    
    for id in delete_rows:
        mali.deleteEntry( id )

    E.info( "ninput=%i, nremoved=%i, noutput=%i" % (nids, len(delete_rows), mali.getLength()))

##------------------------------------------------------------
##------------------------------------------------------------
##------------------------------------------------------------
def maskColumns( mali, annotation_type, annotation_file ):
    """mask columns in mali according to annotation type."""

    if annotation_type.upper() == "SLR":
    
        try:
            import WrapperSlr
        except ImportError:
            print "# could not add Slr annotation: module WrapperSlr not found."
            return
        
        slr = WrapperSlr.Slr()
        
        result = slr.parseOutput( open(annotation_file, "r").readlines() )

        columns = []

        for site in result.mSites:
            if site.isPositive():
                columns += list( range( 3 * (site.mResidue - 1), 3 * (site.mResidue - 1) + 3) )

    mali.maskColumns( columns )

##------------------------------------------------------------
##------------------------------------------------------------
##------------------------------------------------------------
def addAnnotation(mali, annotation_type, annotation_file ):
    """add annotation to a multiple alignment."""

    if annotation_type.upper() in ("SLR-CODON", "SLR-AA"):
        
        try:
            import WrapperSlr
        except ImportError:
            print "# could not add Slr annotation: module WrapperSlr not found."
            return
        
        slr = WrapperSlr.Slr()
        
        result = slr.parseOutput( open(annotation_file, "r").readlines() )

        codes = []

        if annotation_type.upper() == "SLR-CODON":
            l = 3
        else:
            l = 1
        
        for site in result.mSites:
            if site.mResult:
                if site.mResult == "0":
                    c = "@@@"[:l]
                else:
                    c = site.mResult[:l]
            else:
                c = ""
                
            codes.append( c + "_" * (l - len(c)) )

        mali.addAnnotation( "slr", "".join(codes))

    elif annotation_type.upper() in ("CODEML-CODON", "CODEML-AA"):

        try:
            import WrapperCodeML
        except ImportError:
            print "# could not add Codeml annotation: module WrapperCodeML not found."
            return
        
        codeml = WrapperCodeML.CodeMLSites()
        
        result = codeml.parseOutput( open(annotation_file, "r").readlines() )

        codes = []

        if annotation_type.upper() == "CODEML-CODON":
            l = 3
        else:
            l = 1

        threshold = 0.95
        if options.loglevel >= 2:
            options.stdlog.write("# Warning: only positive sites of > %f significance for Bayes Empirical Bayes\n" % threshold)
            
        for x in ("1", "0", "3", "2", "7", "8"):

            sites = result.mSites[x].mBEB.mPositiveSites
            if len(sites) == 0: continue
            
            codes = ["-"] * mali.getWidth()
            found = False
            for site in sites:
                if site.mProbability < threshold: continue
                found = True
                pos = (site.mResidue - 1) * l
                codes[pos:pos+l] = ["+"] * l
            if not found: continue
            mali.addAnnotation( "codeml_%s_%s" % (x, "beb"), "".join(codes))

##------------------------------------------------------------
def maskMali( mali, method="seg" ):
    """mask multiple alignment according to an external masker.
    """
    
    if method == "seg":
        masker = Masker.MaskerSeg()
    elif method == "bias":
        masker = Masker.MaskerBias()
    elif method == "random":
        masker = Masker.MaskerRandom()
        
    if mali.getAlphabet() == "na" and method in ("seg", "bias"):
        for id, s in mali.items():
            ss = Genomics.TranslateDNA2Protein( s.mString )
            mss = masker( ss )
            columns = []
            for x in range( 0, len(mss) ):
                if mss[x] in ("X", "x"):
                    columns += range( x, x+3 )
            mali.getEntry(id).maskColumns( columns )
    else:
        for id, s in mali.items():
            mali[id].mString = masker( s.mString )

##------------------------------------------------------------
def filterMali( mali, method="3rd" ):
    """build a new multiple alignment based on a filter.

    valid methods are
    3rd:        only third positions
    4d:         only four-fold degenerate sites
    """

    if method not in ("3rd", "4d"):
        raise "unknown method %s" % method

    if method == "3rd":
        columns = range( 2, mali.getWidth(), 3 )
        
    elif method == "4d":
        ## translate
        trans_mali = Mali.Mali()
        for id, seq in mali.items():
            s = []
            sequence = seq.mString
            l = len(sequence)
            for codon in [ sequence[x:x+3] for x in range(0, l, 3) ]:
                aa = Genomics.MapCodon2AA( codon )
                s.append( aa )

            trans_mali.addSequence( id, 0, l, "".join( s ) )

        ## get four-fold (or higher) degenerate amino acids            
        aa_columns = trans_mali.getColumns()
        columns = []
        for c in range(len(aa_columns)):
            chars = set(aa_columns[c])
            chars = chars.difference(set(mali.mGapChars))
            if len(chars) == 1:
                char = list(chars)[0].upper()
                try:
                    deg = Genomics.DegeneracyAA[char]
                except KeyError:
                    continue
                if deg >= 4:
                    columns.append( c * 3 )
                    
    mali.takeColumns( columns )
        
##------------------------------------------------------------
def main( argv = sys.argv ):

    parser = E.OptionParser( version = "%prog version: $Id: mali2mali.py 2782 2009-09-10 11:40:29Z andreas $", 
                                    usage = globals()["__doc__"])

    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices=("plain", "fasta", "clustal", "stockholm", "phylip" ),
                      help="input format of multiple alignment [default=%default]."  )
    
    parser.add_option("-o", "--output-format", dest="output_format", type="choice",
                      choices=("plain", "fasta", "stockholm", "phylip", "nexus", "plain-fasta" ),
                      help="output format of multiple alignment [default=%default]."  )

    parser.add_option("--with-ranges", dest="with_ranges", action="store_true",
                      help="output alignment ranges (suffix /from-to after identifier) [default=%default]."  )

    parser.add_option( "--without-ranges", dest="with_ranges", action="store_false",
                      help="do not output alignment ranges (suffix /from-to after identifier) [default=%default]."  )

    parser.add_option( "-u", "--allow-duplicates", dest="allow_duplicates", action="store_true",
                      help="permit duplicate entries [default=%default]."  )

    parser.add_option("-m", "--method", dest="methods", type="string",
                      help="""methods to apply. Several methods can be specified in a ','-separated list [default=%default]."""  )

    parser.add_option("-p", "--parameters", dest="parameters", type="string",
                      help="parameter stack for methods that require one [default=%default]."  )

    parser.add_option("-a", "--mask-char", dest="mask_char", type="string",
                      help="character to identify/set masked characters [default=%default]."  )

    parser.set_defaults(
        input_format="fasta",
        output_format="fasta",
        methods = "",
        parameters = "",
        mask_char = "x",
        gap_chars = "-.nN",
        with_ranges = True,
        allow_duplicates = False,
        )

    (options, args) = E.Start( parser )

    options.methods = options.methods.split(",")
    options.parameters = options.parameters.split(",")    
    
    ## 1. read multiple alignment in various formats
    if options.allow_duplicates:
        mali = Mali.SequenceCollection()
    else:
        mali = Mali.Mali()

    t1 = time.time()
    
    mali.readFromFile( options.stdin, format = options.input_format )

    E.info( "read mali with %i entries in %i seconds." % (len(mali), time.time() - t1))
        
    if len(mali) == 0:
        raise ValueError("empty multiple alignment")

    for method in options.methods:

        t1 = time.time()
        
        if method == "remove-unaligned-ends":
            mali.removeUnalignedEnds()
        elif method == "remove-end-gaps":
            mali.removeEndGaps()
        elif method == "remove-all-gaps":
            mali.removeGaps( minimum_gaps = len(mali) )
        elif method == "remove-any-gaps":
            mali.removeGaps( minimum_gaps = 1 )
        elif method == "remove-some-gaps":
            minimum_gaps = int(options.parameters[0])
            del options.parameters[0]
            mali.removeGaps( minimum_gaps = minimum_gaps )
        elif method == "remove-empty-sequences":
            mali.removeEmptySequences()
        elif method == "upper":
            mali.upperCase()            
        elif method == "lower":
            mali.lowerCase()
        elif method == "mark-codons":
            mali.markCodons()
        elif method == "remove-stops":
            mali.removePattern( lambda x: x.upper() in ("TAG", "TAA", "TGA"),
                                allowed_matches = 0,
                                minimum_matches = 1,
                                delete_frame = 3,
                                search_frame = 3)
        elif method == "shift-alignment":
            map_id2offset = IOTools.ReadMap( open( options.parameters[0], "r"),
                                             map_functions=(str,int) )
            del options.parameters[0]
            mali.shiftAlignment( map_id2offset )
        elif method == "propagate-masks":
            mali.propagateMasks( mask_char = options.mask_char )

        elif method == "recount":
            mali.recount()

        elif method in ("mark-transitions", "filter-odd-transitions", "filter-even-transitions",
                        "keep-even-segments", "keep-odd-segments" ):

            if os.path.exists(options.parameters[0]):
                map_id2transitions = IOTools.readMultiMap( open( options.parameters[0], "r"),
                                                           map_functions=(str,int) )
            else:
                map_id2transitions={}
                r = map(int, options.parameters[0].split(':'))
                r.sort()
                map_id2transitions["mali"] = r

            del options.parameters[0]
            if method == "mark-transitions":
                mali.markTransitions( map_id2transitions )            
            elif method in ("filter-odd-transitions", "keep-even-segments"):
                mali.markTransitions( map_id2transitions, mode = "keep-odd" )
            elif method in ("filter-even-transitions", "keep-odd-segments"):
                mali.markTransitions( map_id2transitions, mode = "keep-even" )                                            

        elif method == "propagate-transitions":
            mali.propagateTransitions()
                
        elif method == "map-annotation":
            ## map annotations in one mali (stockholm-format) to the annotations in another.
            ## Note: the first two sequence identifiers must be shared and the sequence of the
            ## same length
            other_mali = Mali.Mali()
            other_mali.readFromFile( open(options.parameters[0], "r"), format="stockholm")
            del options.parameters[0]            
            mali.copyAnnotations( other_mali )

        elif method == "add-annotation":
            annotation_type, annotation_file = options.parameters[:2]
            del options.parameters[:2]
            AddAnnotation( mali, annotation_type, annotation_file )

        elif method == "mask-columns":
            annotation_type, annotation_file = options.parameters[:2]
            del options.parameters[:2]
            maskColumns( mali, annotation_type, annotation_file )

        elif method == "remove-unaligned-pairs":
            removeUnalignedPairs( mali, options )

        elif method == "filter-3rd":
            filterMali( mali, "3rd" )

        elif method == "filter-4d":
            filterMali( mali, "4d" )

        elif method in ("mask-seg", "mask-bias" ):
            a, b = method.split("-")
            maskMali( mali, b )
            
        elif method == "exclude-with-stop":
            mali.filter( method = "with-stop" )

        elif method == "exclude-with-stop":
            mali.filter( method = "with-frameshift" )

        E.info( "applied method %s in %i seconds." % (method, time.time() - t1))

            
    mali.writeToFile( options.stdout, 
                      format = options.output_format, 
                      write_ranges = options.with_ranges )
        
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv ) )
    
